from types import FunctionType, LambdaType, MethodType
from typing import Iterable
import re


lambda_in_oper = {
    'in': 'in',
    'notin': 'notin'
}

lambda_map_oper = {
    'map': 'map'
}

lambda_regex_oper = {
    're': 're'
}

lambda_query_oper = {
    'gt': '>',
    'eq': '==',
    'ne': '!=',
    'lt': '<',
    'ge': '>=',
    'le': '<=',
    None: '==',
    **lambda_regex_oper
}

lambda_oper = {
    **lambda_query_oper,
    **lambda_in_oper,
    **lambda_map_oper
}

lambda_when = {
    'before': 'before',
    'after': 'after'
}

lambda_handle = {
    'lower': lambda x: str(x).lower(),
    'upper': lambda x: str(x).upper(),
}
lambda_handle['lower'].__name__ = 'lower'
lambda_handle['upper'].__name__ = 'upper'


class Operator(object):

    def __init__(self, field_name, oper, value, when=None, handle=None):
        self.field_name = field_name
        self.oper = oper
        self.value = value
        self.when = when
        self.handle = handle

    def __call__(self, df):
        df = self.when_handle(df, 'before')
        return self.when_handle(self.do_oper(df), 'after')

    def do_oper(self, df):
        raise NotImplementedError

    def when_handle(self, df, when):
        if self.when and self.when == when:
            df[self.field_name] = df[self.field_name].map(self.handle)
        return df

    def __str__(self):

        return '<%s field=> %s, oper=> %s, value=> %s, when=%s, handle=%s>' % (
            self.__class__.__name__,
            self.field_name,
            self.oper,
            self.value,
            self.when,
            self.handle and self.handle.__name__
        )


class EmptyOper(Operator):

    def __init__(self, field_name, oper, value, when=None, handle=None):
        pass

    def __call__(self, df):
        return df

    def do_oper(self, df):
        return df


class MapOper(Operator):

    def __init__(self, field_name, oper, value, when=None, handle=None):
        assert isinstance(value, (
            FunctionType,
            LambdaType,
            MethodType,
            dict
        )), TypeError()
        super().__init__(field_name, oper, value, when, handle)
        self.field_name = field_name
        self.oper = oper
        self.value = value

    def do_oper(self, df):
        df[self.field_name] = df[self.field_name].map(self.value)
        return df


class InOper(Operator):

    def __init__(self, field_name, oper, value, when=None, handle=None):
        assert type(value) != str and isinstance(value, Iterable), TypeError()
        super().__init__(field_name, oper, value, when, handle)
        self.field_name = field_name
        self.oper = oper
        self.value = value

    def do_oper(self, df):
        query = df[self.field_name].isin(self.value)
        if self.oper == 'notin':
            return df[~query].reset_index(drop=True).copy()
        return df[query].reset_index(drop=True).copy()


class QueryOper(Operator):

    def __init__(self, field_name, oper, value, when=None, handle=None):
        assert isinstance(value, (str, int, float)), TypeError()
        super().__init__(field_name, oper, value, when, handle)
        self.field_name = field_name
        self.oper = lambda_query_oper[oper]
        self.value = value

    def do_oper(self, df):
        if self.oper == 're':
            return df[
                df[self.field_name].map(
                    lambda x:bool(re.search(self.value, x, re.I | re.S))
                )
            ].reset_index(drop=True).copy()

        if type(self.value) is str:
            self.value = '"' + self.value.strip('"').strip("'") + '"'
        else:
            self.value = str(self.value)
        query_lambda = '%s %s %s' % (
            self.field_name,
            self.oper,
            self.value
        )
        return df.query(query_lambda).reset_index(drop=True).copy()


def make_query(field_name, oper, value, when=None, handle=None):
    if oper in lambda_in_oper:
        return InOper(field_name, oper, value, when, handle)
    if oper in lambda_map_oper:
        return MapOper(field_name, oper, value, when, handle)
    if oper in lambda_query_oper:
        return QueryOper(field_name, oper, value, when, handle)
    return EmptyOper(field_name, oper, value, when, handle)



# -*- coding:utf-8 -*-

import enum
import pandas as pd

from .field import GXFColmuns
from .operator import make_query
from .util import (
    convert_to_gff,
    convert_to_gtf,
    green,
    red,
    register_method,
    split_by_double_dash
)

REGISTER = '__register'


class GXFType(enum.Enum):
    GTF = 0
    GFF = 1
    UNKNOWN = 3


class GXFReader:
    columns = GXFColmuns()
    strand = {
        '-': -1,
        '+': 1,
        1: '+',
        -1: '-'
    }

    def __init__(self, filename, **kwargs):
        self.filename = filename
        self.kwargs = kwargs
        self.df = self.handle_file()

    def handle_file(self):
        filename = self.filename
        df = pd.read_csv(
            filename,
            header=None,
            names=self.columns,
            skip_blank_lines=True,
            dtype={
                0: str,
                3: int,
                4: int,
            },
            **self.kwargs
        )

        return df


class GXF:
    columns = GXFColmuns()
    gxf_type = None

    def __init__(
        self,
        filename,
        comment='#',
        skiprows=0,
        sep='\t',
        read_cls=None
    ):
        self.comment = comment
        self.skiprows = skiprows
        self.sep = sep
        if read_cls is None:
            read_cls = GXFReader
        elif not isinstance(read_cls, GXFReader):
            raise TypeError('read_cls expected a %s, but got a %s',
                            type(GXFReader), type(read_cls))
        if type(filename) is str:
            self.df = read_cls(filename, **{
                'comment': comment,
                'skiprows': skiprows,
                'sep': sep
            }).df
        elif isinstance(filename, GXFReader):
            self.df = filename.df
        elif isinstance(filename, pd.DataFrame):
            self.df = filename
        else:
            raise TypeError(
                'filename expect a str or GXFReader type, but got a %s' % type(filename))
        if not len(self):
            self.gxf_type = GXFType.UNKNOWN
        else:
            self.gxf_type = GXFType.GFF if self.df.iloc[0, [-1]].map(
                lambda x: '=' in x
            ).any() else GXFType.GTF

    def _do_handle(self, where):
        for col in self.columns:
            handle_name = '%s_handle_%s' % (where, col)

            if hasattr(self, handle_name):
                handle = getattr(self, handle_name)
                self.df[col] = self.df[col].map(handle)

    def _clone(self, filename):
        gxf = GXF(
            filename,
            comment=self.comment,
            skiprows=self.skiprows,
            sep=self.sep
        )
        # # bind method to new GXF class
        # for col in self.columns:
        #     before = 'before_handle_%s' % col
        #     after = 'after_handle_%s' % col
        #     if hasattr(self, before):
        #         setattr(gxf, before, getattr(self, before))
        #     if hasattr(self, after):
        #         setattr(gxf, after, getattr(self, after))
        return gxf

    def filter(self, **kwargs):

        processes = []

        for k, v in kwargs.items():
            field_name, oper, when, handle = split_by_double_dash(k)
            if field_name not in self.columns:
                print()
                print('Only the following field names are supported.')
                for col in self.columns:
                    print('*: ' + green('%s' % col))
                print('But got: %s.' % red(field_name))
                exit(1)

            processes.append(make_query(field_name, oper, v, when, handle))

        # main process
        self._do_handle('before')
        new_df = self.df
        for process in processes:
            new_df = process(new_df)
        self._do_handle('after')


        return self._clone(new_df)

    def __str__(self):
        return str(self.df)

    @property
    def dtypes(self):
        return self.df.dtypes

    def to_dict(self):
        return self.df.to_dict('records')

    def __len__(self):
        return self.df.shape[0]

    def to_gff3(self, filename):
        if self.gxf_type != GXFType.GFF:
            self.df[self.columns.attributes] = self.df[self.columns.attributes].map(
                convert_to_gff)
        self.df.to_csv(
            filename,
            index=None,
            header=None,
            columns=self.columns,
            sep='\t',
        )

    def to_gtf(self, filename):
        if self.gxf_type != GXFType.GTF:
            self.df[self.columns.attributes] = self.df[self.columns.attributes].map(
                convert_to_gtf)
        file = open(filename, 'w', encoding='utf-8')
        for line in self.to_dict():
            line_str = []
            for col in self.columns:
                line_str.append(
                    str(line.get(col, ''))
                )
            file.write('\t'.join(line_str) + '\n')
        file.close()



from setuptools import setup
from setuptools import find_packages

from src.gxf import __author__, __email__, __version__

with open('README.md', 'r', encoding='utf-8') as fp:
    readme = fp.read()

setup(
    name='gxf',
    version=__version__,
    author=__author__,
    author_email=__email__,
    maintainer=__author__,
    maintainer_email=__email__,
    description='A fast gtf/gff parser.',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/dwpeng/gxf',
    keywords=['GFF', 'GTF'],
    packages=find_packages('src'),
    package_dir={"": "src"},
    install_requires=[
        'pandas',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: Implementation',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries'
    ]
)




from .operator import (
    lambda_handle,
    lambda_oper,
    lambda_when
)
from collections import namedtuple


def split_by_double_dash(name: str):
    singel_dash = '_'
    double_dash = '__'
    field_name = oper = when = handle = None
    if double_dash in name:
        name = name.split(double_dash)
        if len(name) == 2:
            if name[1].startswith('after') or name[1].startswith('before'):
                field_name = name[0]
                oper = 'eq'
                when, handle = name[1].split(singel_dash)
            else:
                field_name, oper = name
        if len(name) == 3:
            field_name, oper, when_handle = name
            if singel_dash not in when_handle:
                raise ValueError()
            when, handle = when_handle.split(singel_dash)

        if oper not in lambda_oper:
            raise ValueError()

        if when and when not in lambda_when:
            raise ValueError()

        if handle and handle not in lambda_handle:
            raise ValueError()

    else:
        field_name = name
        oper = 'eq'
        when = None
        handle = None
    Name = namedtuple('name', ['field_name', 'oper', 'when', 'handle'])

    return Name(
        field_name,
        oper,
        lambda_when.get(when),
        lambda_handle.get(handle)
    )


def register_method(self, method_name, method):
    setattr(self, method_name, method)


def green(s):
    return '\033[92m' + str(s) + '\033[0m'


def red(s):
    return '\033[31m' + str(s) + '\033[0m'


def convert_to_gff(x):
    ret = []
    for pair in x.split(';'):
        key, value = pair.split(' ')
        ret.append(
            key + '=' + value.strip('"')
        )
    return ';'.join(ret)


def convert_to_gtf(x):
    ret = []
    for pair in x.split(';'):
        key, value = pair.split('=')
        ret.append(
            key + ' ' + '"' + value.strip('"') + '"'
        )
    return ';'.join(ret)


if __name__ == '__main__':
    name1 = 'type'
    name2 = 'type__eq'
    name3 = 'type__eq__after_lower'
    print(split_by_double_dash(name1).field_name)
    print(split_by_double_dash(name2))
    print(split_by_double_dash(name3))





class GXFColmuns:
    _colmuns = [
        'chr_id',
        'source',
        'type',
        'start',
        'end',
        'score',
        'strand',
        'phase',
        'attributes'
    ]

    def __init__(self):
        for col in self._colmuns:
            self.__dict__[col] = col

    def __len__(self):
        return len(self._colmuns)

    def __iter__(self):
        return iter(self._colmuns)

    @property
    def columns(self):
        return self._colmuns
